package com.dart.cometchat

import com.cometchat.pro.models.*

class GetMap {



    companion object  {
        fun getMessageMap(message: BaseMessage?): HashMap<String, Any?>? {
            if (message == null) return null
            val map: HashMap<String, Any?> = hashMapOf(
                "id" to message.id,
                "muid" to message.muid,
                "sender" to getUserMap(message.sender),
                "receiver" to when (message.receiver) {
                    is User -> getUserMap(message.receiver as User)
                    is Group -> getGroupMap(message.receiver as Group)
                    else -> null
                },
                "receiverUid" to message.receiverUid,
                "type" to message.type,
                "receiverType" to message.receiverType,
                "category" to message.category,
                "sentAt" to message.sentAt,
                "deliveredAt" to message.deliveredAt,
                "readAt" to message.readAt,
                "metadata" to message.metadata?.toString(),
                "readByMeAt" to message.readByMeAt,
                "deliveredToMeAt" to message.deliveredToMeAt,
                "deletedAt" to message.deletedAt,
                "editedAt" to message.editedAt,
                "deletedBy" to message.deletedBy,
                "editedBy" to message.editedBy,
                "updatedAt" to message.updatedAt,
                "conversationId" to message.conversationId,
                "parentMessageId" to message.parentMessageId,
                "replyCount" to message.replyCount
            )
            when (message) {
                is TextMessage -> map["text"] = message.text
                is MediaMessage -> map.putAll(
                    hashMapOf(
                        "caption" to message.caption,
                        "attachment" to getAttachmentMap(message.attachment)
                    )
                )
                is Action -> map.putAll(
                    hashMapOf(
                        "message" to message.message,
                        "rawData" to message.rawData,
                        "action" to message.action,
                        "oldScope" to message.oldScope,
                        "newScope" to message.newScope
                    )
                )
            }
            return map
        }



        private fun getAttachmentMap(attachment: Attachment?): HashMap<String, Any?>? {
            if (attachment == null) return null
            return hashMapOf(
                "fileName" to attachment.fileName,
                "fileExtension" to attachment.fileExtension,
                "fileSize" to attachment.fileSize,
                "fileMimeType" to attachment.fileMimeType,
                "fileUrl" to attachment.fileUrl
            )
        }

        private fun getConversationMap(conversation: Conversation): HashMap<String, Any?> {
            return hashMapOf(
                "conversationId" to conversation.conversationId,
                "conversationType" to conversation.conversationType,
                "conversationWith" to when (conversation.conversationWith) {
                    is User -> getUserMap(conversation.conversationWith as User)
                    is Group -> getGroupMap(conversation.conversationWith as Group)
                    else -> null
                },
                "lastMessage" to getMessageMap(conversation.lastMessage),
                "unreadMessageCount" to conversation.unreadMessageCount,
                "updatedAt" to conversation.updatedAt
            )
        }

        private fun getGroupMap(group: Group): HashMap<String, Any?> {
            val groupMap = group.toMap()
            return hashMapOf(
                "guid" to group.guid,
                "name" to group.name,
                "type" to groupMap["type"],
                "password" to group.password,
                "icon" to group.icon,
                "description" to group.description,
                "owner" to group.owner,
                "metadata" to group.metadata?.toString(),
                "createdAt" to group.createdAt,
                "updatedAt" to group.updatedAt,
                "hasJoined" to (groupMap["hasJoined"]?.toInt() == 1),
                "joinedAt" to group.joinedAt,
                "scope" to group.scope,
                "membersCount" to group.membersCount,
                "tags" to group.tags
            )
        }

        private fun getUserMap(user: User): HashMap<String, Any?> {
            return hashMapOf(
                "uid" to user.uid,
                "name" to user.name,
                "avatar" to user.avatar,
                "link" to user.link,
                "role" to user.role,
                "metadata" to user.metadata?.toString(),
                "status" to user.status,
                "statusMessage" to user.statusMessage,
                "lastActiveAt" to user.lastActiveAt,
                "tags" to user.tags,
                "blockedByMe" to user.isBlockedByMe,
                "hasBlockedMe" to user.isHasBlockedMe
            )
        }

        private fun getGroupMemberMap(groupMember: GroupMember): HashMap<String, Any?> {
            val map: HashMap<String, Any?> = getUserMap(groupMember)
            map.putAll(
                hashMapOf(
                    "scope" to groupMember.scope,
                    "joinedAt" to groupMember.joinedAt
                )
            )
            return map
        }


    }




}