import Flutter
import UIKit
import CometChatPro

public class SwiftCometchatPlugin: NSObject, FlutterPlugin {
    
    var sink: FlutterEventSink?
    //Initialize over here in order to get callback from Cometchat
    //var messagesRequest = MessagesRequest.MessageRequestBuilder().set(limit: 50).build(); // for messages obj
    
    var messagesRequest : MessagesRequest? = nil
    
    var convRequest : ConversationRequest?=nil //for conversation obj
    
    var groupInstance = Group(guid: "", name: "", groupType: .public, password: "");
    
    var groupMembersRequest:GroupMembersRequest?=nil
    
    var groupsRequest  : GroupsRequest?=nil
    
    var mediaMessage = MediaMessage(receiverUid: "", fileurl:"", messageType: .image, receiverType: .user);
    
    var blockedUserRequest = BlockedUserRequest.BlockedUserRequestBuilder().build();
    
    var usersRequest : UsersRequest?=nil
    
    var bannedGroupMembersRequest  : BannedGroupMembersRequest?=nil
    
    public static func register(with registrar: FlutterPluginRegistrar) {
        let instance = SwiftCometchatPlugin()
        
        let channel = FlutterMethodChannel(name: "cometchat", binaryMessenger: registrar.messenger())
        registrar.addMethodCallDelegate(instance, channel: channel)
        
        let channelName = "cometchat_message_stream"
        let eventChannel = FlutterEventChannel(name: channelName, binaryMessenger: registrar.messenger())
        eventChannel.setStreamHandler(instance)
        
        CometChat.messagedelegate = instance
        CometChat.userdelegate = instance
        CometChat.groupdelegate  = instance
        CometChat.logindelegate = instance
        CometChat.connectiondelegate = instance
    }
    
    public func handle(_ call: FlutterMethodCall, result: @escaping FlutterResult) {
        //    result("iOS " + UIDevice.current.systemVersion)
        
        DispatchQueue.main.async { [weak self] in
            let args = call.arguments as? [String: Any] ?? [String: Any]();
            switch call.method {
            case "init":
                self?.initializeCometChat(args: args, result:result)
            case "createUser":
                self?.createUser(args: args, result:result)
            case "loginWithApiKey":
                self?.loginWithApiKey(args: args, result:result)
            case "loginWithAuthToken":
                self?.loginWithAuthToken(args: args, result:result)
            case "logout":
                self?.logout(result:result)
            case "getLoggedInUser":
                self?.getLoggedInUser(result: result)
            case "getUser":
                self?.getUser(args: args, result: result)
            case "sendMessage":
                self?.sendMessage(args: args, result:result)
            case "sendMediaMessage":
                self?.sendMediaMessage(args: args, result:result)
            case "fetchPreviousMessages":
                self?.fetchPreviousMessages(args: args, result:result)
            case "fetchNextConversations":
                self?.fetchNextConversations(args: args, result:result)
            case "deleteMessage":
                self?.deleteMessage(args: args, result:result)
            case "createGroup":
                self?.createGroup(args: args, result:result)
            case "joinGroup":
                self?.joinGroup(args: args, result:result)
            case "leaveGroup":
                self?.leaveGroup(args: args, result:result)
            case "deleteGroup":
                self?.deleteGroup(args: args, result:result)
            case "fetchNextGroupMembers":
                self?.fetchNextGroupMembers(args: args, result:result)
            case "fetchNextGroups":
                self?.fetchNextGroups(args: args, result:result)
            case "registerTokenForPushNotification":
                self?.registerTokenForPushNotification(args: args, result:result)
            case "getUnreadMessageCount":
                self?.getUnreadMessageCount(result:result)
            case "markAsRead":
                self?.markAsRead(args: args, result:result)
            case "callExtension":
                self?.callExtension(args: args, result:result)
            case "blockUsers":
                self?.blockUsers(args: args, result:result)
            case "unblockUsers":
                self?.unblockUsers(args: args, result:result)
            case "fetchBlockedUsers":
                self?.fetchBlockedUsers(result:result)
            case "fetchUsers":
                self?.fetchUsers(args: args, result:result)
            case "startTyping":
                self?.startTyping(args: args, result:result)
            case "endTyping":
                self?.endTyping(args: args, result:result)
            case "getConversation":
                self?.getConversation(args: args, result:result)
            case "fetchNextMessages":
                self?.fetchNextMessages(args: args, result:result)
            case "getUnreadMessageCountForGroup":
                self?.getUnreadMessageCountForGroup(args: args, result:result)
            case "getUnreadMessageCountForAllUsers":
                self?.getUnreadMessageCountForAllUsers(args: args, result:result)
            case "getUnreadMessageCountForAllGroups":
                self?.getUnreadMessageCountForAllGroups(args: args, result:result)
            case "markAsDelivered":
                self?.markAsDelivered(args: args, result:result)
            case "getMessageReceipts":
                self?.getMessageReceipts(args: args, result:result)
            case "editMessage":
                self?.editMessage(args: args, result:result)
            case "deleteConversation":
                self?.deleteConversation(args: args, result:result)
            case "sendTransientMessage":
                self?.sendTransientMessage(args: args, result:result)
            case "getOnlineUserCount":
                self?.getOnlineUserCount(args: args, result:result)
            case "updateUser":
                self?.updateUser(args: args, result:result)
            case "updateCurrentUserDetails":
                self?.updateCurrentUserDetails(args: args, result:result)
            case "getGroup":
                self?.getGroup(args: args, result:result)
            case "getOnlineGroupMemberCount":
                self?.getOnlineGroupMemberCount(args: args, result:result)
            case "updateGroup":
                self?.updateGroup(args: args, result:result)
            case "addMembersToGroup":
                self?.addMembersToGroup(args: args, result:result)
            case "kickGroupMember":
                self?.kickGroupMember(args: args, result:result)
            case "banGroupMember":
                self?.banGroupMember(args: args, result:result)
            case "unbanGroupMember":
                self?.unbanGroupMember(args: args, result:result)
            case "updateGroupMemberScope":
                self?.updateGroupMemberScope(args: args, result:result)
            case "transferGroupOwnership":
                self?.transferGroupOwnership(args: args, result:result)
            case "fetchBlockedGroupMembers":
                self?.fetchBlockedGroupMembers(args: args, result:result)
            case "tagConversation":
                self?.tagConversation(args: args, result:result)
            case "connect":
                self?.connect(result:result)
            case "disconnect":
                self?.disconnect(result:result)
            case "getLastDeliveredMessageId":
                self?.getLastDeliveredMessageId(args: args,result:result)
            case "getConnectionStatus":
                self?.getConnectionStatus(result:result)
                
            default:
                result(FlutterMethodNotImplemented)
            }
        }
    }
    
    
    private func initializeCometChat(args: [String: Any], result: @escaping FlutterResult){
        let appId = args["appId"] as! String
        let region = args["region"] as! String
        let subscriptionType = args["subscriptionType"] as! String ?? "allUsers"
        let roles = args["roles"] as? [String] ?? []
        let autoEstablishSocketConnection = args["autoEstablishSocketConnection"] as? Bool ?? true
        
        var builder = AppSettings.AppSettingsBuilder()
        if (subscriptionType != nil) {
            switch(subscriptionType) {
               case "none"  :
                print("none")
                break;
               case "allUsers"  :
                print("allUsers")
                builder = builder.subscribePresenceForAllUsers()
                break;
               case "roles"  :
                print("allUsers")
                builder = builder.subcribePresenceForRoles(roles: roles)
                break;
               case "roles"  :
                 builder = builder.subcribePresenceForRoles(roles: roles)
                 break;
               default :
                builder = builder.subscribePresenceForAllUsers()
            }
        }
        
        builder = builder.setRegion(region: region)
        builder = builder.autoEstablishSocketConnection(autoEstablishSocketConnection)
        
        //let mySettings = AppSettings.AppSettingsBuilder().subscribePresenceForAllUsers().setRegion(region: region).build()
        
        let mySettings = builder.build()
        CometChat.init(appId: appId ,appSettings: mySettings,onSuccess: { (isSuccess: Bool) in
            print("CometChat Pro SDK intialise successfully.")
            result(String(isSuccess))
        }) { (error) in
            print("CometChat Pro SDK failed intialise with error: \(error.errorDescription)")
            result(FlutterError(code: error.errorCode,
                                message: error.errorDescription, details: error.debugDescription))
        }
    }
    
    private func createUser(args: [String: Any], result: @escaping FlutterResult){
        let apiKey = args["apiKey"] as! String
        let uid = args["uid"] as! String
        let name = args["name"] as! String
        let avatar = args["avatar"] as! String?
        
        let newUser : User = User(uid: uid, name: name)
        newUser.avatar = avatar
        CometChat.createUser(user: newUser, apiKey: apiKey, onSuccess: { (user) in
            print("User created successfully. \(user.stringValue())")
            result(self.getUserMap(user: user))
        }) { (error) in
            print("The error is \(String(describing: error?.description))")
            result(FlutterError(code: error?.errorCode ?? "",
                                message: error?.errorDescription, details: error?.debugDescription))
        }
    }
    
    private func loginWithApiKey(args: [String: Any], result: @escaping FlutterResult){
        let uid = args["uid"] as! String
        let apiKey = args["apiKey"] as! String
        
        CometChat.login(UID: uid, apiKey: apiKey) { (user) in
            print("User logged in successfully. \(user.stringValue())")
            result(self.getUserMap(user: user))
        } onError: { (error) in
            print("The error is \(String(describing: error.description))")
            result(FlutterError(code: error.errorCode,
                                message: error.errorDescription, details: error.debugDescription))
        }
    }
    
    private func loginWithAuthToken(args: [String: Any], result: @escaping FlutterResult){
        let authToken = args["authToken"] as! String
        
        CometChat.login(authToken: authToken) { (user) in
            print("User logged in successfully. \(user.stringValue())")
            result(self.getUserMap(user: user))
        } onError: { (error) in
            print("The error is \(String(describing: error.description))")
            result(FlutterError(code: error.errorCode,
                                message: error.errorDescription, details: error.debugDescription))
        }
    }
    
    private func logout(result: @escaping FlutterResult){
        CometChat.logout { (message) in
            result(String(message))
        } onError: { (error) in
            result(FlutterError(code: error.errorCode,
                                message: error.errorDescription, details: error.debugDescription))
        }
    }
    
    private func getLoggedInUser(result: @escaping FlutterResult){
        let user = CometChat.getLoggedInUser()
        print("Current User. \(user?.stringValue() ?? "Null")")
        result(getUserMap(user: user))
    }
    
    private func getUser(args: [String: Any], result: @escaping FlutterResult){
        let uid = args["uid"] as? String ?? ""
        
        CometChat.getUser(UID: uid) { (user) in
            result(self.getUserMap(user: user));
        } onError: { (error) in
            print("Error getting user." )
            result(FlutterError(code: error?.errorCode ?? "",
                                message: error?.errorDescription, details: error?.debugDescription))
        }
        
    }
    func convertToDictionary(text: String) -> [String: Any]? {
        if let data = text.data(using: .utf8) {
            do {
                return try JSONSerialization.jsonObject(with: data, options: []) as? [String: Any]
            } catch {
                print(error.localizedDescription)
            }
        }
        return nil
    }

    
    private func sendMessage(args: [String: Any], result: @escaping FlutterResult){
        let receiverID = args["receiverId"] as? String ?? ""
        let messageText = args["messageText"] as? String ?? ""
        let receiver = args["receiverType"] as? String ?? ""
        let parentMessageId = args["parentMessageId"] as? Int ?? 0
        let tags = args["tags"] as? [String] ?? []
        let metadata = args["metadata"] as? Dictionary<String,Any> ?? [:]
        //print("metadata ", metadata)
        
                
        let receiverType : CometChat.ReceiverType
        switch receiver {
        case "user":
            receiverType =  CometChat.ReceiverType.user
        default:
            receiverType =   CometChat.ReceiverType.group
        }
        
        let textMessage = TextMessage(receiverUid: receiverID , text: messageText, receiverType: receiverType)
        

        if parentMessageId > 0 {
            textMessage.parentMessageId = parentMessageId
        }
        
        if (!tags.isEmpty){
            textMessage.tags = tags
        }
        
        if (!metadata.isEmpty){
            textMessage.metaData = metadata
        }
        
        CometChat.sendTextMessage(message: textMessage, onSuccess: { (message) in
            print("TextMessage sent successfully. " + message.stringValue())
            let user = CometChat.getLoggedInUser();
            message.sender = user
            result(self.getMessageMap(message: message))
        }) { (error) in
            print("TextMessage sending failed with error: " + error!.errorDescription);
            result(FlutterError(code: error?.errorCode ?? "",
                                message: error?.errorDescription, details: error?.debugDescription))
        }
    }
    
    private func sendMediaMessage(args: [String: Any], result: @escaping FlutterResult){
        
        let receiverid = args["receiverId"] as? String ?? ""
        let receiver = args["receiverType"] as? String ?? ""
        let msgType = args["messageType"] as? String ?? ""
        
        var filePath:String? = args["filePath"] as? String ?? ""
        let caption = args["caption"] as? String ?? ""
        let parentMessageId = args["parentMessageId"] as? Int ?? 0
        let tags = args["tags"] as? [String] ?? []
        let metadata = args["metadata"] as? Dictionary<String,Any> ?? [:]
        let hasAttachment = args["hasAttachment"] as? Bool ?? false
        let attachmentFileName = args["attachmentFileName"] as? String ?? ""
        let attchemntFileExtension = args["attchemntFileExtension"] as? String ?? ""
        let attachmentFileUrl = args["attachmentFileUrl"] as? String ?? ""
        let attachmentMimeType = args["attachmentMimeType"] as? String ?? ""
        
        
        let messageType : CometChat.MessageType
        switch msgType {
        case "image":
            messageType = .image
        case "video":
            messageType = .video
        case "audio":
            messageType = .audio
        default:
            messageType = .file
        }
        
        let receiverType : CometChat.ReceiverType
        switch receiver {
        case "user":
            receiverType =  CometChat.ReceiverType.user
        default:
            receiverType =   CometChat.ReceiverType.group
        }
        
        if(hasAttachment){
            filePath = nil
        }
        
        
        let mediaMessage = MediaMessage(receiverUid: receiverid, fileurl:filePath, messageType:  messageType, receiverType: receiverType)
        if(hasAttachment){
            mediaMessage.attachment = Attachment(fileName: attachmentFileName, fileExtension: attchemntFileExtension, fileMimeType: attachmentMimeType, fileUrl: attachmentFileUrl);
        }
        
        
        if (caption != ""){
            mediaMessage.caption = caption
        }
        
        if (parentMessageId > 0) {
            mediaMessage.parentMessageId = parentMessageId
        }
        
        if(!tags.isEmpty){
            mediaMessage.tags = tags
        }
        if (!metadata.isEmpty){
            mediaMessage.metaData = metadata
        }
        
        CometChat.sendMediaMessage(message: mediaMessage, onSuccess: { (response) in
            let user = CometChat.getLoggedInUser();
            response.sender = user
            result(self.getMessageMap(message: response))
        }) { (error) in
            result(FlutterError(code: error?.errorCode ?? "",
                                message: error?.errorDescription, details: error?.debugDescription))
        }
    }
    
    private func fetchNextConversations(args: [String: Any], result: @escaping FlutterResult){
        
        let limit = args["limit"] as? Int ?? 50
        let typeValue = args["type"] as? String
        let initializeBuilder = args["init"] as? Bool ?? true
        let withUserAndGroupTags = args["withUserAndGroupTags"] as? Bool ?? true
        let withTags = args["withTags"] as? Bool ?? true
        let tags = args["tags"] as? [String] ?? []
        
        
        if(initializeBuilder || self.convRequest==nil){
            self.convRequest = ConversationRequest.ConversationRequestBuilder(limit: 20).setConversationType(conversationType: .user).build()
            var builder = ConversationRequest.ConversationRequestBuilder(limit: limit)
            
            if let typeValue = typeValue {
                if typeValue == "user" {
                    builder = builder.setConversationType(conversationType: .user)
                } else if typeValue == "group" {
                    builder = builder.setConversationType(conversationType: .group)
                }
            }
            if(withUserAndGroupTags){
                builder.withUserAndGroupTags(withUserAndGroupTags)
            }
            if(withTags){
                builder = builder.withTags(true)
            }
            if(!tags.isEmpty){
                builder = builder.setTags(tags)
            }
           
        self.convRequest = builder.build()
        }
        
            //      If we want to fetch by Conversations
            
        
        
        self.convRequest?.fetchNext(onSuccess: { (conversationList) in
            let list = conversationList.map { (e) -> [String : Any]? in
                return self.getConversationMap(conversation: e)
            }
            print("list count ",list.count)
            result(list)
        }) { (exception) in
            print("here exception \(String(describing: exception?.errorDescription))")
            result(FlutterError(code: exception?.errorCode ?? "",
                                message: exception?.errorDescription, details: exception?.debugDescription))
        }
        
        //        If we Want to fetch by Messages
        //        self.messagesRequest = MessagesRequest.MessageRequestBuilder().set(limit: limit).build();
        //        self.messagesRequest.fetchNext(onSuccess: { (response) in
        //            print("Message count is ,", response?.count ?? 0)
        //
        //            if let messages = response{
        //                let conversationList = messages.map { (eachMsg) -> [String : Any]? in
        //                    if let conversation = CometChat.getConversationFromMessage(eachMsg){
        //                        return self.getConversationMap(conversation: conversation)
        //                    }
        //                    return [:]
        //                }
        //                result(conversationList)
        //            }
        //
        //        }) { (error) in
        //
        //          print("Message receiving failed with error: " + error!.errorDescription);
        //          print("here exception \(String(describing: error?.errorDescription))")
        //          result(FlutterError(code: error?.errorCode ?? "",
        //                                message: error?.errorDescription, details: error?.debugDescription))
        //        }
    }
    
    private func fetchPreviousMessages(args: [String: Any], result: @escaping FlutterResult){
        let limit = args["limit"] as? Int ?? 50
        let uid = args["uid"] as? String ?? ""
        let guid = args["guid"] as? String ?? ""
        let searchTerm = args["searchTerm"] as? String ?? ""
        let messageId = args["messageId"] as? Int ?? 0
        
        let timestamp = args["timestamp"] as? Int ?? 0
        let unread = args["unread"] as? Bool ?? false
        let hideblockedUsers = args["hideblockedUsers"] as? Bool ?? false
        let updateAfter = args["updateAfter"] as? Int ?? 0
        let updatesOnly = args["updatesOnly"] as? Bool ?? false
        let categories = args["categories"] as? [String] ?? []
        let types = args["types"] as? [String] ?? []
        let parentMessageId = args["parentMessageId"] as? Int ?? 0
        let hideReplies = args["hideReplies"] as? Bool ?? false
        let hideDeletedMessages = args["hideDeletedMessages"] as? Bool ?? false
        let initializeBuilder = args["init"] as? Bool ?? true
        let withTags = args["withTags"] as? Bool ?? true
        let tags = args["tags"] as? [String] ?? []
        
        if(initializeBuilder || self.messagesRequest==nil){
            self.messagesRequest = MessagesRequest.MessageRequestBuilder().set(limit: 50).build()
            var builder = MessagesRequest.MessageRequestBuilder()
            builder = builder.set(categories: ["message"])

            if (limit > 0) {
                builder = builder.set(limit: limit)
            }
            
            
            if (uid != "") {
                builder = builder.set(uid: uid)
            } else if (guid != "") {
                builder = builder.set(guid: guid)
            }
            if (searchTerm != ""){
                builder = builder.set(searchKeyword: searchTerm)
            }
            if (messageId > 0){
                builder = builder.set(messageID: messageId)
            }
            
            
            if (timestamp > 0) {
                builder = builder.set(timeStamp: timestamp)
            }
            if (unread) {
                builder = builder.set(unread:unread)
            }
            if (hideblockedUsers) {
                builder = builder.hideMessagesFromBlockedUsers(hideblockedUsers)
            }
            if (updateAfter > 0) {
                builder = builder.setUpdatedAfter(timeStamp: updateAfter)
            }
            if (updatesOnly) {
                builder = builder.updatesOnly(onlyUpdates: updatesOnly)
            }
            if (!categories.isEmpty){
                builder = builder.set(categories: categories)
            }
            if (!types.isEmpty){
                builder = builder.set(types: types)
            }
            if (parentMessageId > 0) {
                builder = builder.setParentMessageId(parentMessageId: parentMessageId)
            }
            if (hideReplies) {
                builder = builder.hideReplies(hide: hideReplies)
            }
            if (hideDeletedMessages) {
                builder = builder.hideDeletedMessages(hide: hideDeletedMessages)
            }
            if(withTags){
                builder = builder.withTags(withTags)
            }
            if(!tags.isEmpty){
                builder = builder.setTags(tags)
            }
            
            
            builder = builder.set(timeStamp: timestamp)
            
            self.messagesRequest = builder.build()
            
        }
        
        
        self.messagesRequest?.fetchPrevious(onSuccess: { (response) in
            if let messages = response{
                let conversationList = messages.map { (eachMsg) -> [String : Any]? in
                    return self.getMessageMap(message: eachMsg)
                    
                }
                result(conversationList)
            }
        }) { (error) in
            print("Message receiving failed with error: " + error!.errorDescription);
            result(FlutterError(code: error?.errorCode ?? "" ,
                                message: error?.errorDescription, details: error?.debugDescription))
        }
        
    }
    private func deleteMessage(args: [String: Any], result: @escaping FlutterResult){
        let msgID = args["messageId"] as? Int ?? 1
        
        CometChat.delete(messageId: msgID, onSuccess: { (baseMessage) in
            //print("message deleted successfully. \(baseMessage)")
            result(nil)
        }) { (error) in
            //print("delete message failed with error: \(error.errorDescription)")
            result(FlutterError(code: error.errorCode ,
                                message: error.errorDescription, details: error.debugDescription))
        }
        
    }
    private func createGroup(args: [String: Any], result: @escaping FlutterResult){
        
        let guid = args["guid"] as? String ?? ""
        let groupName = args["groupName"] as? String ?? ""
        let grpType = args["groupType"] as? String ?? ""
        let groupType : CometChat.groupType = grpType == "private" ? .private : grpType == "public" ? .public : .password
        let password = args["password"] as? String ?? nil //mandatory in case of password protected group type
        
        self.groupInstance = Group(guid: guid, name: groupName, groupType: groupType, password: password);
        
        CometChat.createGroup(group: self.groupInstance, onSuccess: { (group) in
            
            print("Group created successfully. " + group.stringValue())
            result(self.getGroupMap(group: group))
            
        }) { (error) in
            
            print("Group creation failed with error:" + error!.errorDescription);
            result(FlutterError(code: error?.errorCode ?? "" ,
                                message: error?.errorDescription, details: error?.debugDescription))
        }
        
    }
    private func joinGroup(args: [String: Any], result: @escaping FlutterResult){
        
        let guid = args["guid"] as? String ?? ""
        let grpType = args["groupType"] as? String ?? ""
        let groupType : CometChat.groupType = grpType == "private" ? .private : grpType == "public" ? .public : .password
        let password = args["password"] as? String ?? nil //mandatory in case of password protected group type
        
        CometChat.joinGroup(GUID: guid, groupType: groupType, password: password, onSuccess: { (group) in
            
            print("Group joined successfully. " + group.stringValue())
            result(self.getGroupMap(group: group))
            
        }) { (error) in
            
            print("Group joining failed with error:" + error!.errorDescription);
            result(FlutterError(code: error?.errorCode ?? "" ,
                                message: error?.errorDescription, details: error?.debugDescription))
        }
        
    }
    private func leaveGroup(args: [String: Any], result: @escaping FlutterResult){
        
        let guid = args["guid"] as? String ?? ""
        CometChat.leaveGroup(GUID: guid, onSuccess: { (response) in
            
            print("Left group successfully.")
            result(response)
            
        }) { (error) in
            
            print("Group leaving failed with error:" + error!.errorDescription);
            result(FlutterError(code: error?.errorCode ?? "" ,
                                message: error?.errorDescription, details: error?.debugDescription))
        }
        
    }
    private func deleteGroup(args: [String: Any], result: @escaping FlutterResult){
        
        let guid = args["guid"] as? String ?? ""
        CometChat.deleteGroup(GUID: guid, onSuccess: { (response) in
            
            print("Group deleted successfully.")
            result(response)
            
        }) { (error) in
            
            print("Group delete failed with error: " + error!.errorDescription);
            result(FlutterError(code: error?.errorCode ?? "" ,
                                message: error?.errorDescription, details: error?.debugDescription))
        }
        
    }
    private func fetchNextGroupMembers(args: [String: Any], result: @escaping FlutterResult){
        
        let limit = args["limit"] as? Int ?? 50
        let guid = args["guid"] as? String ?? ""
        let keyword = args["keyword"] as? String ?? ""
        let initializeBuilder = args["init"] as? Bool ?? true
        let scopes = args["scopes"] as? [String] ?? []
        
        
        if(initializeBuilder || self.groupMembersRequest==nil){
            self.groupMembersRequest = GroupMembersRequest.GroupMembersRequestBuilder(guid: "").set(limit: 50).build();
            var builder = GroupMembersRequest.GroupMembersRequestBuilder(guid: guid)
            if (limit > 0) {
                builder = builder.set(limit: limit)
            }
            
            if (keyword != ""){
                builder = builder.set(searchKeyword: keyword)
            }
            if(!scopes.isEmpty){
                builder = builder.set(scopes: scopes)
            }
            
            self.groupMembersRequest = builder.build();
            
        }
        
        groupMembersRequest?.fetchNext(onSuccess: { (groupMembers) in
            
            result(self.getGroupMemberMap(users: groupMembers))
            
        }) { (error) in
            
            print("Group Member list fetching failed with error:" + error!.errorDescription);
            result(FlutterError(code: error?.errorCode ?? "" ,
                                message: error?.errorDescription, details: error?.debugDescription))
        }
        
    }
    private func fetchNextGroups(args: [String: Any], result: @escaping FlutterResult){
        
        let limit = args["limit"] as? Int ?? 50
        let searchTerm = args["searchTerm"] as? String ?? ""
        let initializeBuilder = args["init"] as? Bool ?? true
        let joinedOnly = args["joinedOnly"] as? Bool ?? true
        let tags = args["tags"] as? [String] ?? []
        let withTags = args["withTags"] as? Bool ?? false
        
        
        if(initializeBuilder || self.groupsRequest==nil){
            self.groupsRequest  = GroupsRequest.GroupsRequestBuilder(limit: 50).build();
            var builder = GroupsRequest.GroupsRequestBuilder(limit: limit)
            
            if (searchTerm != "") {
                builder = builder.set(searchKeyword:searchTerm)
            }
            
            if(joinedOnly){
                builder = builder.set(joinedOnly: joinedOnly)
            }
            
            if (!tags.isEmpty){
                builder = builder.set(tags: tags)
            }
            if (withTags){
                builder = builder.withTags(withTags)
            }
            
            self.groupsRequest  = builder.build();
        }
        
        
        
        groupsRequest?.fetchNext(onSuccess: { (groups) in
            
            let list = groups.map { (eachGrp) -> [String : Any]? in
                return self.getGroupMap(group: eachGrp)
            }
            result(list)
            
        }) { (error) in
            
            print("Groups list fetching failed with error:" + error!.errorDescription);
            result(FlutterError(code: error?.errorCode ?? "" ,
                                message: error?.errorDescription, details: error?.debugDescription))
        }
        
    }
    
    private func getUnreadMessageCount(result: @escaping FlutterResult){
        CometChat.getUnreadMessageCount(onSuccess: { (response) in
            print("Unread message count: \(response)")
            result(response)
            
        }) { (error) in
            print("Error in fetching unread count: \(error)")
            result(FlutterError(code: error?.errorCode ?? "" ,
                                message: error?.errorDescription, details: error?.debugDescription))
        }
    }
    
    private func markAsRead(args: [String: Any], result: @escaping FlutterResult){
        let baseMessageMap = args["baseMessage"] as? Dictionary<String,AnyObject?> ?? [:]
    
        let baseMessage = toBaseMessage(messageMap: baseMessageMap)
        print("________________________________")
        print(baseMessage.id )
        print("Receaver id : "+(baseMessage.receiver as! User).uid! )
        print("Receaver name : "+(baseMessage.receiver as! User).name! )
        //dump(baseMessageMap)
        print("receaver  Uid : "+(baseMessage.receiverUid))
        print("sender  Uid : "+(baseMessage.senderUid))
        print(baseMessage.receiverType)

        CometChat.markAsRead(  baseMessage: baseMessage, onSuccess: {
            print("markAsRead Succces")
            result("Success")
        }, onError: {(error) in
            result(FlutterError(code: error?.errorCode ?? "" ,
                                message: error?.errorDescription, details: error?.debugDescription))
        })
    }
    
    private func callExtension(args: [String: Any], result: @escaping FlutterResult){
        
        let slug = args["slug"] as? String ?? ""
        let requestType = args["requestType"] as? String ?? ""
        let postType : HTTPMethod = requestType == "POST" ? .post : .get
        let endPoint = args["endPoint"] as? String ?? ""
        let body = args["body"] as? [String:Any]
        
        CometChat.callExtension(slug: slug,
                                type: postType,
                                endPoint: endPoint,
                                body: body) { (response) in
//            print("Response: \(response)")
            result(response)
            
        } onError: { (error) in
            
            result(FlutterError(code: error?.errorCode ?? "" ,
                                message: error?.errorDescription, details: error?.debugDescription))
        }
    }
    
    private func registerTokenForPushNotification(args: [String: Any], result: @escaping FlutterResult){
        let token = args["token"] as? String ?? ""
        CometChat.registerTokenForPushNotification(token: token) { (response) in
            result(nil)
        } onError: { (error) in
            result(FlutterError(code: error?.errorCode ?? "" ,
                                message: error?.errorDescription, details: error?.debugDescription))
        }
    }
    
    private func blockUsers(args: [String: Any], result: @escaping FlutterResult){
        let uids = args["uids"] as? [String] ?? []
        CometChat.blockUsers(uids, onSuccess: { (users) in
            print("Blocked user successfully.")
            result(users)
        }, onError: { (error) in
            print("Blocked user failed with error: \(error?.errorDescription)")
            result(FlutterError(code: error?.errorCode ?? "" ,
                                message: error?.errorDescription, details: error?.debugDescription))
        })
    }
    
    private func unblockUsers(args: [String: Any], result: @escaping FlutterResult){
        let uids = args["uids"] as? [String] ?? []
        
        CometChat.unblockUsers(uids) { (users) in
            print("Unblocked user successfully.")
            result(users)
        } onError: { (error) in
            print("Unblocked user failed with error: \(error?.errorDescription)")
            result(FlutterError(code: error?.errorCode ?? "" ,
                                message: error?.errorDescription, details: error?.debugDescription))
        }
    }
    
    private func fetchBlockedUsers(result: @escaping FlutterResult){
        blockedUserRequest = BlockedUserRequest.BlockedUserRequestBuilder().build();
        blockedUserRequest.fetchNext (onSuccess: { (blockedUsers) in
            let users = blockedUsers ?? []
            
            let list = users.map({ (e) -> [String : Any]? in
                return self.getUserMap(user: e)
            })
            result(list)
            
        }, onError: { (error) in
            print("exco Error while fetching the blocked user request:  \(error?.errorDescription)");
            result(FlutterError(code: error?.errorCode ?? "" ,
                                message: error?.errorDescription, details: error?.debugDescription))
        })
    }
    
    private func fetchUsers(args: [String: Any], result: @escaping FlutterResult){
        
        let searchTerm = args["searchTerm"] as? String ?? ""
        let limit = args["limit"] as? Int ?? 50
        let hidebloackedUsers = args["hidebloackedUsers"] as? Bool ?? false
        let userRoles = args["userRoles"] as? [String] ?? []
        let friendsOnly = args["friendsOnly"] as? Bool ?? false
        let tags = args["tags"] as? [String] ?? []
        let uids = args["uids"] as? [String] ?? []
        let withTags = args["withTags"] as? Bool ?? false
        let userStatus = args["userStatus"] as? String ?? ""
        print("userStatus" + String(userStatus))
        NSLog("userStatus", String(userStatus) )
        
        let initializeBuilder = args["init"] as? Bool ?? true
        
        
        if(initializeBuilder || self.groupsRequest==nil){
            
            usersRequest = UsersRequest.UsersRequestBuilder().set(limit: 50).build();
            
            var builder = UsersRequest.UsersRequestBuilder();
            
            if (limit > 0) {
                builder = builder.set(limit: limit)
            }
            if (searchTerm != "") {
                builder = builder.set(searchKeyword:searchTerm)
            }
            if (hidebloackedUsers){
                builder = builder.hideBlockedUsers(hidebloackedUsers)
            }
            if (!userRoles.isEmpty){
                builder = builder.set(roles: userRoles)
            }
            if (friendsOnly){
                builder = builder.friendsOnly( friendsOnly)
            }
            if (!tags.isEmpty){
                builder = builder.set(tags: tags)
            }
            if (!uids.isEmpty){
                builder = builder.set(UIDs:uids)
            }
            if (withTags){
                builder = builder.withTags(withTags)
            }
            if (userStatus != "") {
                if userStatus.compare("online", options: .caseInsensitive) == .orderedSame {
                    print("Setting it online")
                    builder = builder.set(status:CometChat.UserStatus.online)
                }
                else{
                    print("Setting it offline")
                    builder = builder.set(status: CometChat.UserStatus.offline)
                }
                
            }
            
            self.usersRequest = builder.build();
            
            
        }
        
        
        usersRequest?.fetchNext(onSuccess: { (userList) in
            let list = userList.map { (e) -> [String : Any]? in
                return self.getUserMap(user: e)
            }
            //print("list  count ",list.count)
            result(list)
            
        }) { (error) in
            
            print("Users list fetching failed with error:" + error!.errorDescription);
            result(FlutterError(code: error?.errorCode ?? "" ,
                                message: error?.errorDescription, details: error?.debugDescription))
        }
        
    }
    
    
    
    private func startTyping(args: [String: Any], result: @escaping FlutterResult){
        
        let uid = args["uid"] as? String ?? ""
        let receiver = args["receiverType"] as? String ?? ""
        let receiverType : CometChat.ReceiverType
        switch receiver {
        case "user":
            receiverType =  CometChat.ReceiverType.user
        default:
            receiverType =   CometChat.ReceiverType.group
        }
        let typingIndicator = TypingIndicator(receiverID: uid, receiverType: receiverType)
        CometChat.startTyping(indicator: typingIndicator)
        
    }
    private func endTyping(args: [String: Any], result: @escaping FlutterResult){
        
        let uid = args["uid"] as? String ?? ""
        let receiver = args["receiverType"] as? String ?? ""
        let receiverType : CometChat.ReceiverType
        switch receiver {
        case "user":
            receiverType =  CometChat.ReceiverType.user
        default:
            receiverType =   CometChat.ReceiverType.group
        }
        let typingIndicator = TypingIndicator(receiverID: uid, receiverType: receiverType)
        CometChat.endTyping(indicator: typingIndicator)
        
    }
    
    private func getConversation(args: [String: Any], result: @escaping FlutterResult){
        
        let conversationWith = args["conversationWith"] as? String ?? ""
        let conversationType = args["conversationType"] as? String ?? ""
        let passedConversationType:CometChat.ConversationType
        switch conversationType {
        case "user":
            passedConversationType =  CometChat.ConversationType.user
        default:
            passedConversationType =   CometChat.ConversationType.group
        }
        
        CometChat.getConversation(conversationWith: conversationWith, conversationType: passedConversationType, onSuccess: { (conversation) in
            print("success \(String(describing: conversation?.stringValue()))")
            result(self.getConversationMap(conversation: conversation))
        }) { (exception) in
            print("here exception \(String(describing: exception?.errorDescription))")
            result(FlutterError(code: exception?.errorCode ?? "",
                                message: exception?.errorDescription, details: exception?.debugDescription))
        }
        
    
    }
    
    private func fetchNextMessages(args: [String: Any], result: @escaping FlutterResult){
        let limit = args["limit"] as? Int ?? 50
        let uid = args["uid"] as? String ?? ""
        let guid = args["guid"] as? String ?? ""
        let searchTerm = args["searchTerm"] as? String ?? ""
        let messageId = args["messageId"] as? Int ?? 0
        
        let timestamp = args["timestamp"] as? Int ?? 0
        let unread = args["unread"] as? Bool ?? false
        let hideblockedUsers = args["hideblockedUsers"] as? Bool ?? false
        let updateAfter = args["updateAfter"] as? Int ?? 0
        let updatesOnly = args["updatesOnly"] as? Bool ?? false
        let categories = args["categories"] as? [String] ?? []
        let types = args["types"] as? [String] ?? []
        let parentMessageId = args["parentMessageId"] as? Int ?? 0
        let hideReplies = args["hideReplies"] as? Bool ?? false
        let hideDeletedMessages = args["hideDeletedMessages"] as? Bool ?? false
        let withTags = args["withTags"] as? Bool ?? false
        let tags = args["tags"] as? [String] ?? []
        let initializeBuilder = args["init"] as? Bool ?? true
        
        
        if(initializeBuilder || self.messagesRequest==nil){
            self.messagesRequest = MessagesRequest.MessageRequestBuilder().set(limit: 50).build()
            var builder = MessagesRequest.MessageRequestBuilder()
            builder = builder.set(categories: ["message"])

            if (limit > 0) {
                builder = builder.set(limit: limit)
            }
            
            
            if (uid != "") {
                builder = builder.set(uid: uid)
            } else if (guid != "") {
                builder = builder.set(guid: guid)
            }
            if (searchTerm != ""){
                builder = builder.set(searchKeyword: searchTerm)
            }
            if (messageId > 0){
                builder = builder.set(messageID: messageId)
            }
            
            
            if (timestamp > 0) {
                builder = builder.set(timeStamp: timestamp)
            }
            if (unread) {
                builder = builder.set(unread:unread)
            }
            if (hideblockedUsers) {
                builder = builder.hideMessagesFromBlockedUsers(hideblockedUsers)
            }
            if (updateAfter > 0) {
                builder = builder.setUpdatedAfter(timeStamp: updateAfter)
            }
            if (updatesOnly) {
                builder = builder.updatesOnly(onlyUpdates: updatesOnly)
            }
            if (!categories.isEmpty){
                builder = builder.set(categories: categories)
            }
            if (!types.isEmpty){
                builder = builder.set(types: types)
            }
            if (parentMessageId > 0) {
                builder = builder.setParentMessageId(parentMessageId: parentMessageId)
            }
            if (hideReplies) {
                builder = builder.hideReplies(hide: hideReplies)
            }
            if (hideDeletedMessages) {
                builder = builder.hideDeletedMessages(hide: hideDeletedMessages)
            }
            if(withTags){
                builder = builder.withTags(withTags)
            }
            if(!tags.isEmpty){
                builder = builder.setTags(tags)
            }
                
            
            
            builder = builder.set(timeStamp: timestamp)
            
            self.messagesRequest = builder.build()
            
        }
        
        
        self.messagesRequest?.fetchNext(onSuccess: { (response) in
            if let messages = response{
                let conversationList = messages.map { (eachMsg) -> [String : Any]? in
                    return self.getMessageMap(message: eachMsg)
                    
                }
                result(conversationList)
            }
        }) { (error) in
            print("Message receiving failed with error: " + error!.errorDescription);
            result(FlutterError(code: error?.errorCode ?? "" ,
                                message: error?.errorDescription, details: error?.debugDescription))
        }
        
    }
    
    
    private func getUnreadMessageCountForGroup(args: [String: Any],result: @escaping FlutterResult){
        
        let guid = args["guid"] as? String ?? ""
        let hideMessagesFromBlockedUsers = args["hideMessagesFromBlockedUsers"] as? Bool ?? false
        
        CometChat.getUnreadMessageCountForGroup(guid,hideMessagesFromBlockedUsers: hideMessagesFromBlockedUsers, onSuccess: { (response) in
            print("Unread message count: \(response)")
            result(response)
            
        }) { (error) in
            print("Error in fetching unread count: \(error)")
            result(FlutterError(code: error?.errorCode ?? "" ,
                                message: error?.errorDescription, details: error?.debugDescription))
        }
    }
    
    private func getUnreadMessageCountForAllUsers(args: [String: Any],result: @escaping FlutterResult){
        let hideMessagesFromBlockedUsers = args["hideMessagesFromBlockedUsers"] as? Bool ?? false
        
        CometChat.getUnreadMessageCountForAllUsers(hideMessagesFromBlockedUsers: hideMessagesFromBlockedUsers, onSuccess: { (response) in
            print("Unread message count: \(response)")
            result(response)
            
        }) { (error) in
            print("Error in fetching unread count: \(error)")
            result(FlutterError(code: error?.errorCode ?? "" ,
                                message: error?.errorDescription, details: error?.debugDescription))
        }
    }
    
    
    private func getUnreadMessageCountForAllGroups(args: [String: Any],result: @escaping FlutterResult){
        let hideMessagesFromBlockedUsers = args["hideMessagesFromBlockedUsers"] as? Bool ?? false
        
        
        CometChat.getUnreadMessageCountForAllGroups(hideMessagesFromBlockedUsers: hideMessagesFromBlockedUsers, onSuccess: { (response) in
            print("Unread message count: \(response)")
            result(response)
            
        }) { (error) in
            print("Error in fetching unread count: \(error)")
            result(FlutterError(code: error?.errorCode ?? "" ,
                                message: error?.errorDescription, details: error?.debugDescription))
        }
    }
    
    
    private func markAsDelivered(args: [String: Any], result: @escaping FlutterResult){
        let baseMessageMap = args["baseMessage"] as? Dictionary<String,AnyObject?> ?? [:]
    
        let baseMessage = toBaseMessage(messageMap: baseMessageMap)

        CometChat.markAsDelivered(  baseMessage: baseMessage, onSuccess: {
            print("markAsRead Succces")
            result("Success")
        }, onError: {(error) in
            result(FlutterError(code: error?.errorCode ?? "" ,
                                message: error?.errorDescription, details: error?.debugDescription))
        })
    }
    
    
    private func getMessageReceipts(args: [String: Any], result: @escaping FlutterResult){
        let messageId = args["id"] as? Int ?? -1


        CometChat.getMessageReceipts(messageId, onSuccess: { (receipt) in
                    
          print("getMessageReceipt \(receipt)")
            
        }) { (error) in
          print(error?.errorDescription)
            result(FlutterError(code: error?.errorCode ?? "" ,
                                message: error?.errorDescription, details: error?.debugDescription))
        }
        
    }
    
    
    private func editMessage(args: [String: Any], result: @escaping FlutterResult){
        let baseMessageMap = args["baseMessage"] as? Dictionary<String,AnyObject?> ?? [:]
    
        let baseMessage = toTextMessage(messageMap: baseMessageMap)

        CometChat.edit(message: baseMessage, onSuccess: { (baseMessage) in
                        
            print("success \(String(describing: baseMessage.description))")
            result(self.getMessageMap(message:baseMessage))
                        
        }) { (error) in
            print(error.errorDescription)
              result(FlutterError(code: error.errorCode ?? "" ,
                                  message: error.errorDescription, details: error.debugDescription))
        }
    }
    
    private func deleteConversation(args: [String: Any], result: @escaping FlutterResult){
        
        let conversationWith = args["conversationWith"] as? String ?? ""
        let conversationType = args["conversationType"] as? String ?? ""
        let passedConversationType:CometChat.ConversationType
        switch conversationType {
        case "user":
            passedConversationType =  CometChat.ConversationType.user
        default:
            passedConversationType =   CometChat.ConversationType.group
        }

        CometChat.deleteConversation(conversationWith: conversationWith, conversationType: passedConversationType, onSuccess: { message in
            print("Conversation deleted",message)
            result(message)
        }, onError: {error in
            print("exco Error while fetching the blocked user request:  \(error?.errorDescription)");
            result(FlutterError(code: error?.errorCode ?? "" ,
                                message: error?.errorDescription, details: error?.debugDescription))
        })
        
    }
    
    
    private func sendTransientMessage(args: [String: Any], result: @escaping FlutterResult){
        
        let receiverId = args["receiverId"] as? String ?? ""
        let receiverType = args["receiverType"] as? String ?? ""
        let data = args["data"] as? Dictionary<String,String> ?? [:]
        
        
        let passedReceiverType:CometChat.ReceiverType
        switch receiverType {
        case "user":
            passedReceiverType =  CometChat.ReceiverType.user
        default:
            passedReceiverType =   CometChat.ReceiverType.group
        }

        let transientMessage = TransientMessage(receiverID: receiverId, receiverType: passedReceiverType, data: data)
        CometChat.sendTransientMessage(message: transientMessage)
        result("Message Sent")
        
    }
    
    
    private func getOnlineUserCount(args: [String: Any], result: @escaping FlutterResult){
        
        CometChat.getOnlineUserCount(onSuccess: { count in
            print("Conversation deleted",count)
            result(count)
        }, onError: {error in
            print("exco Error while fetching onlint user count:  \(error?.errorDescription)");
            result(FlutterError(code: error?.errorCode ?? "" ,
                                message: error?.errorDescription, details: error?.debugDescription))
        })
        
    }
    
    
    private func updateUser(args: [String: Any], result: @escaping FlutterResult){
        
        let apiKey = args["apiKey"] as? String ?? ""
        let uid = args["uid"] as? String ?? ""
        let name = args["name"] as? String ?? ""
        
        let updateUser : User = User(uid: uid, name: name) // Replace with your uid and the name for the user to be created.
        CometChat.updateUser(user: updateUser, apiKey: apiKey, onSuccess: { (user) in
             print("User updated successfully. \(user.stringValue())")
            result(self.getUserMap(user: user));
         }) { (error) in
             print("The error is \(String(describing: error?.description))")
             result(FlutterError(code: error?.errorCode ?? "" ,
                                 message: error?.errorDescription, details: error?.debugDescription))
         }
        
    }
    
    
    private func updateCurrentUserDetails(args: [String: Any], result: @escaping FlutterResult){
        
        let uid = args["uid"] as? String ?? ""
        let name = args["name"] as? String ?? ""
        
        let updateUser : User = User(uid: uid, name: name) // Replace with your uid and the name for the user to be created.
        CometChat.updateCurrentUserDetails(user: updateUser,  onSuccess: { (user) in
             print("User updated successfully. \(user.stringValue())")
            result(self.getUserMap(user: user));
         }) { (error) in
             print("The error is \(String(describing: error?.description))")
             result(FlutterError(code: error?.errorCode ?? "" ,
                                 message: error?.errorDescription, details: error?.debugDescription))
         }
        
    }
    
    private func getGroup(args: [String: Any], result: @escaping FlutterResult){
        
        let guid = args["guid"] as? String ?? ""
        
        CometChat.getGroup(GUID: guid, onSuccess: { (group) in
            print("Group details fetched successfully. " + group.stringValue())
            result(self.getGroupMap(group: group));

        }) { (error) in
            print("The error is \(String(describing: error?.description))")
            result(FlutterError(code: error?.errorCode ?? "" ,
                                message: error?.errorDescription, details: error?.debugDescription))
        }
        
    }
    
    private func getOnlineGroupMemberCount(args: [String: Any], result: @escaping FlutterResult){
        
        let guids = args["guids"] as? [String] ?? []
        
        CometChat.getOnlineGroupMemberCount(guids, onSuccess: { countData in
            print(countData)
            result(countData);
        }, onError: { error in
            print("The error is \(String(describing: error?.description))")
            result(FlutterError(code: error?.errorCode ?? "" ,
                                message: error?.errorDescription, details: error?.debugDescription))
         })
        
    }
    private func updateGroup(args: [String: Any], result: @escaping FlutterResult){
        
        let guid = args["guid"] as? String ?? ""
        let groupName = args["groupName"] as? String ?? ""
        let passedGroupType = args["groupType"] as? String ?? ""
        let passedPassword = args["password"] as? String ?? ""
        let groupType : CometChat.groupType = passedGroupType == "private" ? .private : passedGroupType == "public" ? .public : .password
        
        let groupTobeUpdated = Group(guid: guid, name: groupName, groupType: groupType, password: passedPassword)

        CometChat.updateGroup(group: groupTobeUpdated, onSuccess: { (group) in
          print("Groups details updated successfully. " + group.stringValue())
            result(self.getGroupMap(group: group));
            

        }) { (error) in
            print("The error is \(String(describing: error?.description))")
            result(FlutterError(code: error?.errorCode ?? "" ,
                                message: error?.errorDescription, details: error?.debugDescription))
        }
    }
    
    private func addMembersToGroup(args: [String: Any], result: @escaping FlutterResult){
        
        let guid = args["guid"] as? String ?? ""
        let groupMembers = args["groupMembers"] as? [Dictionary<String,AnyObject?>] ?? []
        var groupMemberList = [GroupMember]()
        
        for groupMember in groupMembers {
            groupMemberList.append(toGroupMember(groupMemberMap: groupMember))
        }
        
        
        CometChat.addMembersToGroup(guid: guid, groupMembers: groupMemberList, onSuccess: { (response) in
                print("Response from addMembersGroup: \(response)")
            result(response)
            
        }, onError : { (error) in
            print("The error is \(String(describing: error?.description))")
            result(FlutterError(code: error?.errorCode ?? "" ,
                                message: error?.errorDescription, details: error?.debugDescription))
            
        })
    }
    

    
    private func kickGroupMember(args: [String: Any], result: @escaping FlutterResult){
        
        let guid = args["guid"] as? String ?? ""
        let uid = args["uid"] as? String ?? ""
        
        CometChat.kickGroupMember(UID: uid, GUID: guid, onSuccess: { (response) in

          print("\(uid) is kicked from the group \(guid) successfully.")
            result(response)

        }) { (error) in

            print("The error is \(String(describing: error?.description))")
            result(FlutterError(code: error?.errorCode ?? "" ,
                                message: error?.errorDescription, details: error?.debugDescription))
        }
    }
    private func banGroupMember(args: [String: Any], result: @escaping FlutterResult){
        
        let guid = args["guid"] as? String ?? ""
        let uid = args["uid"] as? String ?? ""
        
        CometChat.banGroupMember(UID: uid, GUID: guid, onSuccess: { (response) in

          print("\(uid) is kicked from the group \(guid) successfully.")
            result(response)

        }) { (error) in

            print("The error is \(String(describing: error?.description))")
            result(FlutterError(code: error?.errorCode ?? "" ,
                                message: error?.errorDescription, details: error?.debugDescription))
        }
    }
    
    private func unbanGroupMember(args: [String: Any], result: @escaping FlutterResult){
        
        let guid = args["guid"] as? String ?? ""
        let uid = args["uid"] as? String ?? ""
        
        CometChat.unbanGroupMember(UID: uid, GUID: guid, onSuccess: { (response) in

          print("\(uid) is kicked from the group \(guid) successfully.")
            result(response)

        }) { (error) in

            print("The error is \(String(describing: error?.description))")
            result(FlutterError(code: error?.errorCode ?? "" ,
                                message: error?.errorDescription, details: error?.debugDescription))
        }
    }
    
    private func updateGroupMemberScope(args: [String: Any], result: @escaping FlutterResult){
        
        let GUID = args["guid"] as? String ?? ""
        let UID = args["uid"] as? String ?? ""
        let passedScope = args["scope"] as? String ?? ""
        let scope : CometChat.MemberScope = passedScope == "admin" ? .admin : passedScope == "moderator" ? .moderator : .participant
        
        CometChat.updateGroupMemberScope(UID: UID, GUID: GUID, scope: scope, onSuccess: { (response) in

             print("Update group member scope changed successfully.")
            result(response)

        }) { (error) in

            print("The error is \(String(describing: error?.description))")
            result(FlutterError(code: error?.errorCode ?? "" ,
                                message: error?.errorDescription, details: error?.debugDescription))
        }
        
    }
    
    
    private func transferGroupOwnership(args: [String: Any], result: @escaping FlutterResult){
        
        let GUID = args["guid"] as? String ?? ""
        let UID = args["uid"] as? String ?? ""
        CometChat.transferGroupOwnership(UID: UID, GUID: GUID) { (response) in
                    
            print("Transfer Group Ownership Successfully")
            result(response)

        } onError: { (error) in

            print("Transfer Group Ownership failed with error \(error?.errorDescription)")
            print("The error is \(String(describing: error?.description))")
            result(FlutterError(code: error?.errorCode ?? "" ,
                                message: error?.errorDescription, details: error?.debugDescription))
        }
        
    }
    
    
    private func fetchBlockedGroupMembers(args: [String: Any], result: @escaping FlutterResult){
        
        let searchTerm = args["searchKeyword"] as? String ?? ""
        let limit = args["limit"] as? Int ?? 50
        let guid = args["guid"] as? String ?? ""
        let initializeBuilder = args["init"] as? Bool ?? true
        
        
        if(initializeBuilder || self.bannedGroupMembersRequest==nil){
            
            //bannedGroupMembersRequest = BannedGroupMembersRequest.BannedGroupMembersRequestBuilder(guid: guid).build();
            
            var builder = BannedGroupMembersRequest.BannedGroupMembersRequestBuilder(guid: guid);
            
            if (limit > 0) {
                builder = builder.set(limit: limit)
            }
            if (searchTerm != "") {
                builder = builder.set(searchKeyword:searchTerm)
            }
            
            self.bannedGroupMembersRequest = builder.build();
            
            
        }
        
        bannedGroupMembersRequest?.fetchNext(onSuccess: { (groupMembers) in
            
            result(self.getGroupMemberMap(users: groupMembers))
            
        }) { (error) in
            
            print("Banned member list fetching failed with error:" + error!.errorDescription);
            result(FlutterError(code: error?.errorCode ?? "" ,
                                message: error?.errorDescription, details: error?.debugDescription))
        }
        
    }
    
    

    private func getLastDeliveredMessageId(args: [String: Any], result: @escaping FlutterResult){
        let lastMessageDeliveredId = CometChat.getLastDeliveredMessageId();
        result(lastMessageDeliveredId)
        
    }
    
    

    private func getConnectionStatus( result: @escaping FlutterResult){
        let connectionStatus = CometChat.getConnectionStatus?.value
        result(connectionStatus)
        
    }
    
    
    
    private func getConversationMap(conversation: Conversation?) -> [String: Any]? {
        if let conversation = conversation {
            var conversationWith : [String : Any]?
            var conversationType : String
            
            switch conversation.conversationType {
            case .user:
                conversationType = "user"
                conversationWith = getUserMap(user: conversation.conversationWith as? User)
            default:
                conversationType = "group"
                conversationWith = getGroupMap(group: conversation.conversationWith as? Group)
            }
            return [
                "conversationId" : conversation.conversationId ?? "",
                "conversationType" : conversationType,
                "conversationWith" : conversationWith ?? [:],
                "updatedAt" : Int(conversation.updatedAt) ,
                "unreadMessageCount" : conversation.unreadMessageCount,
                "lastMessage":self.getMessageMap(message: conversation.lastMessage) ?? [:],
                "tags" : conversation.tags,
            ]
        } else {
            return nil
        }
    }
    
    private func getMessageMap(message: BaseMessage?, methodName: String = "") -> [String: Any]? {
        //print(message as Any)
        if let message = message {
            
            var receiver : [String : Any]?
            var receiverType : String
            switch message.receiverType {
            case .user:
                receiverType = "user"
                receiver = getUserMap(user: message.receiver as? User)
            default:
                receiverType = "group"
                receiver = getGroupMap(group: message.receiver as? Group)
            }
            
            var category : String = "custom"
            switch message.messageCategory {
            case CometChat.MessageCategory.message:
                category = "message"
            case CometChat.MessageCategory.action:
                category = "action"
            case CometChat.MessageCategory.call:
                category = "call"
            default:
                category = "custom"
            }
            
            var type : String = "custom"
            if category != "action"{
                switch message.messageType {
                case CometChat.MessageType.text:
                    type = "text"
                case CometChat.MessageType.image:
                    type = "image"
                case CometChat.MessageType.video:
                    type = "video"
                case CometChat.MessageType.file:
                    type = "file"
                case CometChat.MessageType.audio:
                    type = "audio"
                default:
                    type = "custom"
                }
            }else{
                type = "action"
            }
            
            
            
            var messageMap = [
                "id" : message.id as Any,
                "muid" : message.muid  as Any,
                "sender" : getUserMap(user: message.sender) as Any,
                "receiver" : receiver as Any,
                "receiverUid" : message.receiverUid as Any,
                "type" : type,
                "receiverType" : receiverType,
                "category" : category,
                "sentAt" : Int(message.sentAt),
                "deliveredAt" : Int(message.deliveredAt),
                "readAt" : Int(message.readAt),
                "metadata" :  toJson(dictionary: message.metaData) as Any,
                "readByMeAt" : Int(message.readByMeAt),
                "deliveredToMeAt" : Int(message.deliveredToMeAt),
                "deletedAt" : Int(message.deletedAt),
                "editedAt" : Int(message.editedAt),
                "deletedBy" : message.deletedBy,
                "editedBy" : message.editedBy,
                "updatedAt" : Int(message.updatedAt),
                "conversationId" : message.conversationId,
                "parentMessageId" : message.parentMessageId,
                "replyCount" : message.replyCount,
                "senderUid":  message.senderUid,
                "methodName":methodName
            ]
            
            if let text = message as? TextMessage {
                let map = [
                    "text" : text.text,
                    "tags" : text.tags
                ] as [String : Any]
                map.forEach { (key, value) in messageMap[key] = value }
            } else if let media = message as? MediaMessage{
                let map = [
                    "caption" : media.caption as Any,
                    "attachment" : getAttachmentMap(attachment: media.attachment) as Any,
                    "tags" : media.tags
                ]
                map.forEach { (key, value) in messageMap[key] = value }
            }  else if let customMessage = message as? CustomMessage{
                let map = [
                    "customData" : customMessage.customData as Any,
                    "subType" : customMessage.subType,
                    "tags" : customMessage.tags
                ]
                map.forEach { (key, value) in messageMap[key] = value }
            } else if let action = message as? ActionMessage{
                let map = [
                    "message" : action.message as Any,
                    "rawData" : action.rawData as Any,
                    "action" : action.action?.rawValue as Any,
                    "oldScope" : action.oldScope.rawValue,
                    "newScope" : action.newScope.rawValue,
                ]
                map.forEach { (key, value) in messageMap[key] = value }
            }
            return messageMap;
            
        } else {
            return nil
        }
    }
    
    private func getAttachmentMap(attachment: Attachment?) -> [String: Any]? {
        if let attachment = attachment {
            return [
                "fileName" : attachment.fileName,
                "fileExtension" : attachment.fileExtension,
                "fileSize" : Int(attachment.fileSize) ,
                "fileMimeType" : attachment.fileMimeType as Any,
                "fileUrl" : attachment.fileUrl as Any,
            ]
        } else {
            return nil
        }
    }
    
    private func getUserMap(user: User?, methodName : String = "") -> [String: Any]? {
        if let user = user {
            let status : String
            switch user.status {
            case .online:
                status = "online"
            default:
                status = "offline"
            }
            let userMap = [
                "uid" : user.uid ?? "",
                "name" : user.name ?? "",
                "avatar" : user.avatar ?? "",
                "link" : user.link ?? "",
                "role" : user.role ?? "",
                "metadata" : toJson(dictionary: user.metadata) as Any,
                "status" : status,
                "statusMessage" : user.statusMessage ?? "",
                "lastActiveAt" : Int(user.lastActiveAt) ,
                "tags" : user.tags,
                "blockedByMe" : user.blockedByMe,
                "hasBlockedMe" : user.hasBlockedMe,
                "methodName":methodName
            ]
            return userMap;
        } else {
            return nil
        }
    }
    
    private func getGroupMap(group: Group?, methodName : String = "")-> [String: Any]? {
        let type : String
        switch group?.groupType {
        case .public:
            type = "public"
        case .private:
            type = "private"
        case .password:
            type = "password"
        default:
            type = "public"
        }
        
        if let group = group {
            return [
                "guid" : group.guid ,
                "name" : group.name ?? "",
                "type" : type,
                "password" : group.password ?? "",
                "icon" : group.icon ?? "",
                "description" : group.description ,
                "owner" : group.owner ?? "",
                "metadata" : toJson(dictionary: group.metadata) as Any,
                "createdAt" : group.createdAt ,
                "updatedAt" : group.updatedAt,
                "hasJoined" : group.hasJoined,
                "joinedAt" : group.joinedAt,
                "scope" : group.scope.rawValue,
                "membersCount" : group.membersCount,
                "tags" : group.tags,
                "methodName":methodName
            ]
        } else {
            return nil
        }
    }
    
    private func getGroupMemberMap(users: [GroupMember]) -> [[String:Any]]{
        var usersMap = [[String:Any]]()
        users.forEach { (user) in
            
            let status : String
            switch user.status {
            case .online:
                status = "online"
            default:
                status = "offline"
            }
            let eachUser = [
                "uid" : user.uid ?? "",
                "name" : user.name ?? "",
                "avatar" : user.avatar ?? "",
                "link" : user.link ?? "",
                "role" : user.role ?? "",
                "metadata" : toJson(dictionary: user.metadata) as Any,
                "status" : status,
                "statusMessage" : user.statusMessage ?? "",
                "lastActiveAt" : Int(user.lastActiveAt) ,
                "tags" : user.tags,
                "scope" : user.scope.rawValue,
                "joinedAt" : user.joinedAt
            ]
            usersMap.append(eachUser)
        }
        return usersMap
    }
    
    
    private func getTypingIndicatorMap(typingIndicator: TypingIndicator?, methodName:String="")-> [String: Any]? {
        
        let receiverType : String
        switch typingIndicator?.receiverType{
        case .user:
            receiverType = "user"
        case .group:
            receiverType = "group"
        default:
            receiverType = "user"
        }
        if let typingIndicator = typingIndicator {
            return [
                "sender" : getUserMap(user: typingIndicator.sender) ,
                "receiverId" : typingIndicator.receiverID ,
                "receiverType" : receiverType,
                "metadata" : typingIndicator.metadata ,
                "methodName":methodName
            ]
        } else {
            return nil
        }
    
    }
    
    
    private func getMessageReceiptMap(messageReceipt: MessageReceipt?, methodName:String="")-> [String: Any]? {
        
        
        let receiverType : String
        switch messageReceipt?.receiverType{
        case .user:
            receiverType = "user"
        case .group:
            receiverType = "group"
        default:
            receiverType = "user"
        }
        
        let receiptType : String
        switch messageReceipt?.receiptType{
        case .delivered:
            receiptType = "delivered"
        case .read:
            receiptType = "read"
        default:
            receiptType = "delivered"
        }
        
        print("message receipt is  ", messageReceipt?.messageId)
        
        
        print("message receipt is  ", messageReceipt?.messageId)
        
        if let messageReceipt = messageReceipt {
            return [
                "messageId" : messageReceipt.messageId,
                "sender" : getUserMap(user: messageReceipt.sender) ,
                "receivertype" : receiverType ,
                "receiverId" : messageReceipt.receiverId ,
                "timestamp" : messageReceipt.timeStamp ,
                "receiptType" : receiptType ,
                "deliveredAt" : messageReceipt.deliveredAt ,
                "readAt" : messageReceipt.readAt,
                "methodName":methodName
            ]
        } else {
            return nil
        }
    
    }
    
    private func getTransientMessageMap(transientMessage: TransientMessage?, methodName:String="")-> [String: Any]? {
        let receiverType : String
        switch transientMessage?.receiverType{
        case .user:
            receiverType = "user"
        case .group:
            receiverType = "group"
        default:
            receiverType = "user"
        }
        
        if let transientMessage = transientMessage {
            return [
                "receiverId" : transientMessage.receiverID ,
                "receiverType" : receiverType,
                "data" : transientMessage.data as Any ,
                "methodName":methodName
            ]
        } else {
            return nil
        }
    
    }
    
    private func toJson(dictionary: [String: Any]?)-> String? {
        if let dictionary = dictionary {
            if let jsonData = (try? JSONSerialization.data(withJSONObject: dictionary , options: [])){
                let jsonString = String(data: jsonData, encoding: String.Encoding.ascii)
                return jsonString
            } else {
                return nil
            }
        }else {
            return nil
        }
        
    }
    
    private func tagConversation(args: [String: Any], result: @escaping FlutterResult){
        
        let conversationWith = args["conversationWith"] as? String ?? ""
        let conversationType = args["conversationType"] as? String ?? ""
        let tags = args["tags"] as? [String] ?? []
        let passedConversationType:CometChat.ConversationType
        switch conversationType {
        case "user":
            passedConversationType =  CometChat.ConversationType.user
        default:
            passedConversationType =   CometChat.ConversationType.group
        }
        
        CometChat.tagConversation(conversationWith: conversationWith, conversationType: passedConversationType,
                        tags: tags        ,onSuccess: { (conversation) in
            print("success \(String(describing: conversation?.stringValue()))")
            result(self.getConversationMap(conversation: conversation))
        }) { (exception) in
            print("here exception \(String(describing: exception?.errorDescription))")
            result(FlutterError(code: exception?.errorCode ?? "",
                                message: exception?.errorDescription, details: exception?.debugDescription))
        }
        
    
    }
    
    private func connect( result: @escaping FlutterResult){
        CometChat.connect()
    }
    
    private func disconnect( result: @escaping FlutterResult){
        CometChat.disconnect()
    }
    
}


//MARK: - ------------ Cometchat Message Delegates -----------------
extension SwiftCometchatPlugin : CometChatMessageDelegate{
    
    public func onTextMessageReceived(textMessage: TextMessage) {
        guard let eventSink = self.sink else { return }
        let parsedMsg = self.getMessageMap(message: textMessage,methodName: "onTextMessageReceived")
        print("From textmsg ",parsedMsg)
        eventSink(parsedMsg)
        //return parsedMsg obj back to flutter
    }
    
    public func onMediaMessageReceived(mediaMessage: MediaMessage) {
        guard let eventSink = self.sink else { return }
        let parsedMsg = self.getMessageMap(message: mediaMessage, methodName: "onMediaMessageReceived")
        print("From mediaMsg ",parsedMsg)
        eventSink(parsedMsg)
    }
    
    public func onCustomMessageReceived(customMessage: CustomMessage) {
        guard let eventSink = self.sink else { return }
        let parsedMsg = self.getMessageMap(message: customMessage, methodName: "onCustomMessageReceived")
        print("From textmsg ",parsedMsg)
        eventSink(parsedMsg)
        
    }
    
    public func onTypingStarted(_ typingDetails : TypingIndicator) {
        print("On Typingh Started")
        guard let eventSink = self.sink else { return }
        let parsedIndicator = self.getTypingIndicatorMap(typingIndicator: typingDetails, methodName: "onTypingStarted")
        print("From textmsg ",parsedIndicator)
        eventSink(parsedIndicator)
        print("Typing started  at other end")
    }
    
    public func onTypingEnded(_ typingDetails: TypingIndicator) {
        guard let eventSink = self.sink else { return }
        let parsedIndicator = self.getTypingIndicatorMap(typingIndicator: typingDetails, methodName: "onTypingEnded")
        print("From typing ended ",parsedIndicator)
        eventSink(parsedIndicator)
        print("Typing end at other end")
    }
    
    public func onTransisentMessageReceived(_ message: TransientMessage) {
        guard let eventSink = self.sink else { return }
        let parsedTransientMessage = self.getTransientMessageMap( transientMessage : message, methodName: "onTransisentMessageReceived")
        print("From textmsg ",parsedTransientMessage)
        eventSink(parsedTransientMessage)
        print("Transient Message")
       }
    
    public func onMessagesDelivered(receipt: MessageReceipt) {
        guard let eventSink = self.sink else { return }
        print("Message Receipt",receipt.messageId)
        let parsedMsg = self.getMessageReceiptMap(messageReceipt: receipt, methodName: "onMessagesDelivered")
        print("From textmsg ",parsedMsg)
        eventSink(parsedMsg)
        print("Typing end at other end")
    }
    
    public func onMessagesRead(receipt: MessageReceipt) {
        guard let eventSink = self.sink else { return }
        let parsedMsg = self.getMessageReceiptMap(messageReceipt: receipt, methodName: "onMessagesRead")
        print("From textmsg ",parsedMsg)
        eventSink(parsedMsg)
        print("Typing end at other end")
    }
    
    public func onMessageEdited(message: BaseMessage) {
        guard let eventSink = self.sink else { return }
        let parsedMsg = self.getMessageMap(message: message, methodName: "onMessageEdited")
        print("From textmsg ",parsedMsg)
        eventSink(parsedMsg)
        print("Typing end at other end")
    }
    
    public func onMessageDeleted(message: BaseMessage) {
        guard let eventSink = self.sink else { return }
        let parsedMsg = self.getMessageMap(message: message,methodName: "onMessageDeleted")
        print("From textmsg  ",parsedMsg)
        eventSink(parsedMsg)
        print("Typing end at other end")
    }
    
}

//MARK: - ------------ Cometchat User Delegates -----------------
extension SwiftCometchatPlugin : CometChatUserDelegate{
    
    public func onUserOnline(user: User) {
        print("on User Online ")
        guard let eventSink = self.sink else { return }
        let parsedUser = self.getUserMap(user: user,methodName: "onUserOnline")
        print("From User ",parsedUser)
        eventSink(parsedUser)
        print("on User Online ")
    }
    
    public func onUserOffline(user: User) {
        guard let eventSink = self.sink else { return }
        let parsedUser = self.getUserMap(user: user,methodName: "onUserOffline")
        print("From User ",parsedUser)
        eventSink(parsedUser)
    }
}


//MARK: - ------------ Cometchat Group Delegates -----------------
extension SwiftCometchatPlugin : CometChatGroupDelegate{
    public func onGroupMemberJoined(action: ActionMessage, joinedUser: User, joinedGroup: Group) {
        print("From onGroupMemberJoined ")
        guard let eventSink = self.sink else { return }
        var responseMap: [String: Any?] = [:]
        responseMap["action"] = getMessageMap(message: action)
        responseMap["joinedUser"] = getUserMap(user: joinedUser)
        responseMap["joinedGroup"] = getGroupMap(group:joinedGroup)
        responseMap["methodName"] = "onGroupMemberJoined"
        print("From User ",responseMap)
        eventSink(responseMap)
    }
    
    public func onGroupMemberLeft(action: ActionMessage, leftUser: User, leftGroup: Group) {
        print("From onGroupMemberLeft ")
        guard let eventSink = self.sink else { return }
        var responseMap: [String: Any?] = [:]
        responseMap["action"] = getMessageMap(message: action)
        responseMap["leftUser"] = getUserMap(user: leftUser)
        responseMap["leftGroup"] = getGroupMap(group:leftGroup)
        responseMap["methodName"] = "onGroupMemberLeft"
        print("From User ",responseMap)
        eventSink(responseMap)
    }
    
    public func onGroupMemberKicked(action: ActionMessage, kickedUser: User, kickedBy: User, kickedFrom: Group) {
        print("From onGroupMemberKicked ")
        guard let eventSink = self.sink else { return }
        var responseMap: [String: Any?] = [:]
        responseMap["action"] = getMessageMap(message: action)
        responseMap["kickedUser"] = getUserMap(user: kickedUser)
        responseMap["kickedBy"] = getUserMap(user: kickedBy)
        responseMap["kickedFrom"] = getGroupMap(group:kickedFrom)
        responseMap["methodName"] = "onGroupMemberKicked"
        print("From User ",responseMap)
        eventSink(responseMap)
    }
    
    public func onGroupMemberBanned(action: ActionMessage, bannedUser: User, bannedBy: User, bannedFrom: Group) {
        print("From onGroupMemberBanned ")
        guard let eventSink = self.sink else { return }
        var responseMap: [String: Any?] = [:]
        responseMap["action"] = getMessageMap(message: action)
        responseMap["bannedUser"] = getUserMap(user: bannedUser)
        responseMap["bannedBy"] = getUserMap(user: bannedBy)
        responseMap["bannedFrom"] = getGroupMap(group:bannedFrom)
        responseMap["methodName"] = "onGroupMemberBanned"
        print("From User ",responseMap)
        eventSink(responseMap)
    }
    
    public func onGroupMemberUnbanned(action: ActionMessage, unbannedUser: User, unbannedBy: User, unbannedFrom: Group) {
        print("From onGroupMemberUnbanned ")
        guard let eventSink = self.sink else { return }
        var responseMap: [String: Any?] = [:]
        responseMap["action"] = getMessageMap(message: action)
        responseMap["unbannedUser"] = getUserMap(user: unbannedUser)
        responseMap["unbannedBy"] = getUserMap(user: unbannedBy)
        responseMap["unbannedFrom"] = getGroupMap(group:unbannedFrom)
        responseMap["methodName"] = "onGroupMemberUnbanned"
        print("From User ",responseMap)
        eventSink(responseMap)
    }
    
    public func onGroupMemberScopeChanged(action: ActionMessage, scopeChangeduser: User, scopeChangedBy: User, scopeChangedTo: String, scopeChangedFrom: String, group: Group) {
        print("From onGroupMemberScopeChanged ")
        guard let eventSink = self.sink else { return }
        var responseMap: [String: Any?] = [:]
        responseMap["action"] = getMessageMap(message: action)
        responseMap["updatedUser"] = getUserMap(user: scopeChangeduser)
        responseMap["updatedBy"] = getUserMap(user: scopeChangedBy)
        responseMap["scopeChangedTo"] = scopeChangedTo
        responseMap["scopeChangedFrom"] = scopeChangedFrom
        responseMap["group"] = getGroupMap(group:group)
        responseMap["methodName"] = "onGroupMemberScopeChanged"
        print("From User ",responseMap)
        eventSink(responseMap)
    }
    
    public func onMemberAddedToGroup(action: ActionMessage, addedBy: User, addedUser: User, addedTo: Group) {
        print("From onMemberAddedToGroup ")
        guard let eventSink = self.sink else { return }
        var responseMap: [String: Any?] = [:]
        responseMap["action"] = getMessageMap(message: action)
        responseMap["addedby"] = getUserMap(user: addedBy)
        responseMap["userAdded"] = getUserMap(user: addedUser)
        responseMap["addedTo"] = getGroupMap(group:addedTo)
        responseMap["methodName"] = "onMemberAddedToGroup"
        print("From User ",responseMap)
        eventSink(responseMap)
    }
    
    
    
    
}



//MARK: - ------------ Cometchat Login Delegates -----------------
extension SwiftCometchatPlugin : CometChatLoginDelegate{
    public func onLoginSuccess(user: User) {
        guard let eventSink = self.sink else { return }
        let parsedUser = self.getUserMap(user: user,methodName: "onLoginSuccess")
        print("From User ",parsedUser)
        eventSink(parsedUser)
    }
    
    public func onLoginFailed(error: CometChatException?) {
        guard let eventSink = self.sink else { return }
        let parsedUser = self.getCometChatExceptionMap(error: error , methodName: "onLoginFailed")
        print("From Login Failed ",parsedUser)
        eventSink(parsedUser)
    }
    
    public func onLogoutSuccess() {
        guard let eventSink = self.sink else { return }
        var responseMap: [String: Any?] = [:]
        responseMap["methodName"] = "onLogoutSuccess"
        print("From Logout ",responseMap)
        eventSink(responseMap)
    }
    
    public func onLogoutFailed(error: CometChatException?) {
        guard let eventSink = self.sink else { return }
        let parsedUser = self.getCometChatExceptionMap(error: error , methodName: "onLogoutFailed")
        print("From Logout Failed ",parsedUser)
        eventSink(parsedUser)
    }
    
    
   private func getCometChatExceptionMap(error: CometChatException?, methodName: String  = "") -> [String: Any]? {
        if let error = error {
            return [
                "code" : error.errorCode ,
                "details" : error.description,
                "messsage" : error.errorDescription ,
                "methodName":methodName
            ]
        } else {
            return nil
        }
        
    }
    
}

extension SwiftCometchatPlugin : CometChatConnectionDelegate{
    public func connecting() {
        guard let eventSink = self.sink else { return }
        var responseMap: [String: Any?] = [:]
        responseMap["methodName"] = "connecting"
        print("From connecting ",responseMap)
        eventSink(responseMap)
  }
    
    public func connected() {
        guard let eventSink = self.sink else { return }
        var responseMap: [String: Any?] = [:]
        responseMap["methodName"] = "connected"
        print("From connected ",responseMap)
        eventSink(responseMap)
  }
    
    public func disconnected() {
        guard let eventSink = self.sink else { return }
        var responseMap: [String: Any?] = [:]
        responseMap["methodName"] = "disconnected"
        print("From disconnected ",responseMap)
        eventSink(responseMap)
  }
    
    
    
}



//MARK: - ------------- FlutterEventChannel Delegates -----------------
extension SwiftCometchatPlugin : FlutterStreamHandler{
    
    public func onListen(withArguments arguments: Any?, eventSink events: @escaping FlutterEventSink) -> FlutterError? {
        debugPrint("On Listen Call")
        self.sink = events
        return nil
    }
    
    public func onCancel(withArguments arguments: Any?) -> FlutterError? {
        self.sink = nil
        return nil
    }
    
    
   
    
    private func toBaseMessage(messageMap : [String:Any?]) -> BaseMessage {
        let receiverType : CometChat.ReceiverType
        switch messageMap["receiverType"] as! String {
        case "user":
            receiverType =  CometChat.ReceiverType.user
        default:
            receiverType =   CometChat.ReceiverType.group
        }
        let sender: [String:AnyObject?] = messageMap["sender"] as? Dictionary<String,AnyObject?> ?? [:]
        let receaver:[String:AnyObject?] = messageMap["receiver"] as? Dictionary<String,AnyObject?> ?? [:]
        
        var type : CometChat.MessageType = CometChat.MessageType.text
        
            switch messageMap["type"] as! String {
            case "text" :
                type = CometChat.MessageType.text
            case "image":
                type = CometChat.MessageType.image
            case "video":
                type = CometChat.MessageType.video
            case "file":
                type = CometChat.MessageType.file
            case "audio":
                type = CometChat.MessageType.audio
            default:
                type = CometChat.MessageType.text
            }
        
        var message = BaseMessage(receiverUid: receaver["uid"] as! String, messageType: type, receiverType: receiverType)
        
        message.senderUid = sender["uid"] as! String
        message.id = messageMap["id"] as! Int
        message.sender = User(uid: sender["uid"] as! String,
                              name: sender["name"] as! String)
        
        if(receiverType ==  CometChat.ReceiverType.group){
            message.receiver = Group(guid: receaver["guid"] as! String, name: receaver["guid"] as! String, groupType: CometChat.groupType.public, password: nil)
        }else{
            message.receiver = User(uid: receaver["uid"] as! String,
                                   name: receaver["name"] as! String)
        }
        
        
        
        return message
    }
    
    
    private func toTextMessage(messageMap : [String:Any?]) -> BaseMessage {
        let receiverType : CometChat.ReceiverType
        switch messageMap["receiverType"] as! String {
        case "user":
            receiverType =  CometChat.ReceiverType.user
        default:
            receiverType =   CometChat.ReceiverType.group
        }
        let sender: [String:AnyObject?] = messageMap["sender"] as? Dictionary<String,AnyObject?> ?? [:]
        let receaver:[String:AnyObject?] = messageMap["receiver"] as? Dictionary<String,AnyObject?> ?? [:]
        
        var type : CometChat.MessageType = CometChat.MessageType.text
        
            switch messageMap["type"] as! String {
            case "text" :
                type = CometChat.MessageType.text
            case "image":
                type = CometChat.MessageType.image
            case "video":
                type = CometChat.MessageType.video
            case "file":
                type = CometChat.MessageType.file
            case "audio":
                type = CometChat.MessageType.audio
            default:
                type = CometChat.MessageType.text
            }
        
        
        let message  = TextMessage(receiverUid: receaver["uid"] as! String, text: messageMap["text"] as! String , receiverType: receiverType)
        
        //var message = BaseMessage(receiverUid: receaver["uid"] as! String, messageType: type, receiverType: receiverType)
        
        message.senderUid = sender["uid"] as! String
        message.id = messageMap["id"] as! Int
        message.sender = User(uid: sender["uid"] as! String,
                              name: sender["name"] as! String)
        
        if(receiverType ==  CometChat.ReceiverType.group){
            message.receiver = Group(guid: receaver["guid"] as! String, name: receaver["guid"] as! String, groupType: CometChat.groupType.public, password: nil)
        }else{
            message.receiver = User(uid: receaver["uid"] as! String,
                                   name: receaver["name"] as! String)
        }
        
        message.messageType = type
        message.tags = messageMap["tags"] as! [String]
        
        
        
        return message
    }
    
    private func toGroupMember(groupMemberMap : [String:Any?]) -> GroupMember {
        let passedScope = groupMemberMap["scope"] as? String ?? ""
        let scope : CometChat.GroupMemberScopeType = passedScope == "admin" ? .admin : passedScope == "moderator" ? .moderator : .participant
        let groupMember  = GroupMember(UID: groupMemberMap["uid"] as! String, groupMemberScope: scope)
        return groupMember
    }
    
    
}
