import 'package:cometchat/Exception/CometChatException.dart';
import 'package:cometchat/models/user.dart';
import 'package:cometchat/utils/constants.dart';
import 'package:flutter/services.dart';

class BlockedUsersRequest {
  static final String directionBlockedByMe = "blockedByMe";
  static final String directionHasBlockedMe = "hasBlockedMe";
  static final String directionBoth = "both";

  static final int maxLimit = 100;
  static final int defaultLimit = 30;

  final int? limit;
  final int? token;
  final String? searchKeyword;
  final String? direction;
  bool? init;

  BlockedUsersRequest._builder(BlockedUsersRequestBuilder builder)
      : limit = builder.limit,
        token = builder.token,
        searchKeyword = builder.searchKeyword,
        direction = builder.direction,
        init =  true;

  ///fetch the list of Users that logged-in user have blocked.
  ///
  ///method could throw [PlatformException] with error codes specifying the cause
  ///
  /// See also [Android Documentation](https://www.cometchat.com/docs/android-chat-sdk/users-block-users) [IOS Documentation](https://www.cometchat.com/docs/ios-chat-sdk/users-block-users)
  Future<List<User>> fetchNext(
  {required Function(List<User> userList )? onSuccess,
      required Function(CometChatException excep )? onError}) async {
    try {
      final result = await channel.invokeMethod('fetchBlockedUsers', {
        'limit': this.limit,
        'searchKeyword': this.searchKeyword,
        'direction': this.direction,
        'init': this.init,
      });
      print("Blocked Users: $result");
      final List<User> res = result.map<User>((e) => User.fromMap(e)).toList();
      if(onSuccess!=null){
        onSuccess(res);
      }
      this.init = false;
      return res ;
    } on PlatformException catch (platformException) {
        if(onError!=null){
        onError(CometChatException(platformException.code, platformException.details,platformException.message));
      }
    } catch (e) {
      throw e;

    }
    return [];
  }
}

class BlockedUsersRequestBuilder {
  int? limit;
  int? token;
  String? searchKeyword;
  String? direction;

  BlockedUsersRequestBuilder();

  BlockedUsersRequest build() {
    return BlockedUsersRequest._builder(this);
  }
}
